#!/usr/bin/env python3
import lib as hila5
import random, itertools

class Alice(object):

    def __init__(self):
        self.pk, self.sk = hila5.crypto_kem_keypair()
        self.num_queries = 0

    def pubkey(self):
        return self.pk

    def query(self, ciphertext, guessed_key):
        self.num_queries += 1
        alice_key = hila5.crypto_kem_dec(ciphertext, self.sk)
        return alice_key == guessed_key

alice = Alice()




################################################################################
# oracle query, assuming we got an "evil" private key.
################################################################################

# if c is 1, [1*M,5*M) is 0 and [5*M,1*M) is 1
M = hila5.q // 8

def oracle(target_bit, guess):
    assert abs(hila5.symmrep(hila5.invq(M))) <= 8 # we need M^-1 mod q to be small

    j = (M - guess) * hila5.invq(M) % hila5.q
    cur_secret = [j * x % hila5.q for x in evil_privkey]

    cur_error = [0] * 1024
    cur_error[(1024 - target_bit) % 1024] = pow(-1, target_bit > 0, hila5.q) # x^1024 = -1, so we need to negate the coefficients we got by rotation

    # we compute and decode the ciphertext
    ciphertext, _, bob_secret = hila5.crypto_kem_enc(alice.pubkey(), cur_secret, cur_error)
    bob_pub, safe, reconc, ecc = hila5.deconstruct_ciphertext(ciphertext)

    ########################################

    # mess with the ciphertext
    if 0 not in safe: safe[0] = 0
    reconc[0] = 1
    for i in (0, 16, 32, 49, 80): ecc[i] ^= 1

    ########################################

    # re-encode the ciphertext and recompute the shared secret
    ciphertext = hila5.construct_ciphertext(bob_pub, safe, reconc, ecc)
    bob_secret_1 = bytes([bob_secret[0] | 1]) + bob_secret[1:]
    bob_key_1 = hila5.hash_shared(alice.pubkey(), ciphertext, bob_secret_1)

    # ORACLE QUERY.
    return alice.query(ciphertext, bob_key_1)

################################################################################
# simple binary search for the secret coefficients.
################################################################################

err_range = (-16, 17) # lower and upper bound for psi16 values
def bisect(target_bit):
    a, b = err_range
    while b - a > 1:
        m = (a + b) // 2
        om = oracle(target_bit, m)
        a, b = (a, m) if om else (m, b)
    return a

################################################################################
# run the attack!
################################################################################

common_gen, alice_pub = hila5.deconstruct_pubkey(alice.pubkey())

class Retry(Exception): pass

while True:

    # choose a guess for Bob's evil secret. Pr[good] = 9.9%
    while True:
        i, j = random.randrange(1024), random.randrange(1024)
        if i == j:
            continue
        for ci, cj in itertools.product((-1, 1), (-1, 1)):
            if (ci * alice_pub[i] + cj * alice_pub[j]) % hila5.q == M:
                evil_privkey = [0] * 1024
                evil_privkey[(1024 - i) % 1024] = -ci
                evil_privkey[(1024 - j) % 1024] = -cj
                break
        else: continue
        break

    # recover key bits and hope the result looks reasonable
    try:
        if oracle(0, -16): # definitely bad
            continue
        recovered = []
        for i in range(1024):
            recovered.append(bisect(i))
            print("coefficient \x1b[36m{:4d}\x1b[0m of Alice's secret is \x1b[32m{:3d}\x1b[0m  (\x1b[33m{}\x1b[0m queries so far)".format(i, recovered[-1], alice.num_queries))
            if len(recovered) >= 8 and all(x in (0, -1) for x in recovered):
                raise Retry
    except Retry:
        print('\x1b[33mprobably guessed wrong, retrying...\x1b[0m')
        continue
    except hila5.Error:
        #TODO these errors are probably fixable
        print('\x1b[31msomething blew up...\x1b[0m')
        continue
    break

print("Alice's secret: \x1b[32m{}\x1b[0m".format(recovered))
noise = [hila5.symmrep(x - y) for x, y in zip(alice_pub, hila5.ringmul(common_gen, recovered))]
#print("Alice's noise: {}".format(noise))
if not all(abs(x) <= 16 for x in recovered + noise):
    print('\x1b[31mwrong!\x1b[0m')

sk = hila5.construct_privkey([x % hila5.q for x in recovered], alice.pubkey())
for _ in range(0x10):
    # make sure we can decapsulate with the recovered secret key
    ct, ss, _ = hila5.crypto_kem_enc(alice.pubkey())
    if hila5.crypto_kem_dec(ct, sk) != ss:
        print('\x1b[31mwrong!\x1b[0m')
#import binascii; print("Alice's private key: \x1b[32m{}\x1b[0m".format(binascii.hexlify(sk).decode()))

