#!/usr/bin/env python3
import os.path
import ctypes

q = 12289
symmrep = lambda x: x % q - q if x % q > q // 2 else x % q
invq = lambda x: {1024: q - 12, 3070: 1365, 1536: q - 8}[x % q] # all we need

try:
    lib = ctypes.CDLL(os.path.dirname(__file__) + '/kem.so')
except OSError:
    print("\x1b[31mcouldn't load code from the reference implementation.  see README for instructions.\x1b[0m")
    exit(1)

class Error(Exception): pass

def crypto_kem_keypair():
    pk = ctypes.create_string_buffer(1824)
    sk = ctypes.create_string_buffer(1824)

    if 0 != lib.crypto_kem_keypair(pk, sk):
        raise Error('crypto_kem_keypair returned a non-zero value')

    return bytes(pk), bytes(sk)

def crypto_kem_enc(pubkey, force_secret = None, force_error = None):
    assert len(pubkey) == 1824, 'pubkey has wrong length'

    ct = ctypes.create_string_buffer(2012)
    ss = ctypes.create_string_buffer(32)
    pk = ctypes.create_string_buffer(pubkey)
    fs = fe = ctypes.POINTER(ctypes.c_int32)()
    us = ctypes.create_string_buffer(32) # unhashed secret
    if force_secret: fs = (ctypes.c_int32 * 1024)(*force_secret)
    if force_error: fe = (ctypes.c_int32 * 1024)(*force_error)

    if 0 != lib.crypto_kem_enc(ct, ss, pk, fs, fe, us):
        raise Error('crypto_kem_enc returned a non-zero value')

    return bytes(ct), bytes(ss), bytes(us)

def crypto_kem_dec(ciphertext, privkey):
    assert len(ciphertext) == 2012, 'ciphertext has wrong length'
    assert len(privkey) == 1824, 'privkey has wrong length'

    ss = ctypes.create_string_buffer(32)
    ct = ctypes.create_string_buffer(ciphertext)
    sk = ctypes.create_string_buffer(privkey)

    if 0 != lib.crypto_kem_dec(ss, ct, sk):
        raise Error('crypto_kem_dec returned a non-zero value')

    return bytes(ss)

def pack14(vector):
    assert len(vector) == 1024

    d = ctypes.create_string_buffer(1792)
    v = (ctypes.c_int32 * 1024)(*vector)

    lib.hila5_pack14(d, v)

    return bytes(d)

def unpack14(bs):
    assert len(bs) == 1792

    v = (ctypes.c_int32 * 1024)()
    d = ctypes.create_string_buffer(bs)

    lib.hila5_unpack14(v, d)

    return list(map(int, v))

def ntt(vector, multiplier):
    assert len(vector) == 1024

    d = (ctypes.c_int32 * 1024)()
    v = (ctypes.c_int32 * 1024)(*vector)

    lib.slow_ntt(d, v, ctypes.c_int32(multiplier))

    return list(map(int, d))

def invntt(vector, multiplier):
    assert len(vector) == 1024

    d = (ctypes.c_int32 * 1024)()
    v = (ctypes.c_int32 * 1024)(*vector)

    lib.slow_intt(d, v)

    d = [int(invq(1024 * multiplier) * x % q) for x in d]

    return d

################################################################################

def deconstruct_pubkey(pk):
    seed, packed = pk[:32], pk[32:]
    A = invntt(unpack14(packed), 27)
    G = (ctypes.c_int32 * 1024)()
    lib.hila5_parse(G, ctypes.create_string_buffer(seed))
    G = invntt(G, 1)
    return G, A

def deconstruct_ciphertext(ct):
    if set(ct) == set(b'\0'): # the reference implementation produces this if b=0.
        assert False, 'ciphertext is somehow all zeroes??'

    B, d, c, r = ct[:1792], ct[1792:1920], ct[1920:1982], ct[1982:2012]

    B = invntt(unpack14(B), 27)
    safe = [i for i in range(len(d) * 8) if (int.from_bytes(d, 'little') >> i) & 1]
    reconc = [(int.from_bytes(c, 'little') >> i) & 1 for i in range(len(c) * 8)]
    ecc = [(int.from_bytes(r, 'little') >> i) & 1 for i in range(len(r) * 8)]

    return B, safe, reconc, ecc

def construct_ciphertext(B, safe, reconc, ecc):
    assert len(B) == 1024 and all(type(x) is int for x in B), 'Bob zijn public element is kapot'
    assert len(safe) == len(set(safe)) == 62*8 and all(x in range(1024) for x in safe), 'safe bits zijn kapot'
    assert len(reconc) == 62*8 and all(x in (0, 1) for x in reconc), 'reconciliation info is kapot'
    assert len(ecc) == 30*8 and all(x in (0, 1) for x in ecc), 'error correction is kapot'

    B = pack14(ntt(B, 27))
    safe = sum(1 << i for i in safe).to_bytes(128, 'little')
    reconc = sum(b << i for i, b in enumerate(reconc)).to_bytes(62, 'little')
    ecc = sum(b << i for i, b in enumerate(ecc)).to_bytes(30, 'little')

    ciphertext = B + safe + reconc + ecc
    return ciphertext

def hash_shared(pubkey, ciphertext, unhashed_secret):
    sha3 = ctypes.create_string_buffer(1024) # struct is smaller, but who cares
    h = ctypes.create_string_buffer(32)

    lib.hila5_sha3_init(sha3, 32)

    lib.hila5_sha3_update(sha3, ctypes.create_string_buffer(b'HILA5v10'), 8)

    lib.hila5_sha3(pubkey, len(pubkey), h, len(h))
    lib.hila5_sha3_update(sha3, h, len(h))

    lib.hila5_sha3(ciphertext, len(ciphertext), h, len(h))
    lib.hila5_sha3_update(sha3, h, len(h))

    lib.hila5_sha3_update(sha3, unhashed_secret, len(unhashed_secret))

    lib.hila5_sha3_final(h, sha3)
    return bytes(h)

def construct_privkey(secret, pk):
    sk = pack14(ntt(secret, 27))
    h = ctypes.create_string_buffer(32)
    lib.hila5_sha3(pk, len(pk), h, len(h))
    return sk + bytes(h)

def ringmul(xs, ys):
    return invntt([x * y % q for x, y in zip(ntt(xs, 1), ntt(ys, 1))], 1)

